"use strict";

function _toArray(arr) { return Array.isArray(arr) ? arr : Array.from(arr); }

function refEquals(a, b) {
    return a === b;
}

class Node {
    constructor(children = {}, elements = []) {
        this.children = children;
        this.elements = elements;
    }

    addElements(path, elements) {
        if (path.length === 0) {
            this.elements = [...this.elements, ...elements];
        } else {
            var _path = _toArray(path);

            const first = _path[0],
                  rest = _path.slice(1);

            let child = this.children[first];
            if (!child) {
                child = new Node();
                this.children[first] = child;
            }
            child.addElements(rest, elements);
        }
    }

    addElement(path, element) {
        this.addElements(path, [element]);
    }

    isEmpty() {
        return this.elements.length === 0 && Object.keys(this.children).length === 0;
    }

    removeElement(path, element, compareWith = refEquals) {
        const children = this.children,
              elements = this.elements;

        if (path.length === 0) {
            this.elements = elements.filter(e => !compareWith(e, element));
            return;
        }

        var _path2 = _toArray(path);

        const first = _path2[0],
              rest = _path2.slice(1);

        const nextNode = this.children[first];
        if (this.isEmpty() || !nextNode) {
            return;
        }

        nextNode.removeElement(rest, element, compareWith);

        // Prune child branch from current node if we just removed all descendant elements
        if (nextNode.isEmpty()) {
            delete this.children[first];
        }
    }

    /**
     * Get the node at the given path
     *
     * @param {Array<string>} path
     * @return {Node|null} Node at the given path. Null if node does not exist
     */
    walk(path) {
        if (path.length === 0) {
            return this;
        }

        var _path3 = _toArray(path);

        const first = _path3[0],
              rest = _path3.slice(1);

        const nextNode = this.children[first];
        return nextNode ? nextNode.walk(rest) : null;
    }

    /**
     * Collect all of the elements stored from the node passed down to the element at path
     * 
     * @param {Array<string>} path
     * @return {Array} Collected elements
     */
    collect(path) {
        if (path.length === 0) {
            return this.elements;
        }

        var _path4 = _toArray(path);

        const first = _path4[0],
              rest = _path4.slice(1);

        const nextNode = this.children[first];
        if (!nextNode) {
            return this.elements;
        }

        return [...nextNode.collect(rest), ...this.elements];
    }

    /**
     * Gets the trie of elements matching the prefix path
     *
     * @param {Array<string>} path
     * @returns {Trie|null}
     */
    subTrie(path) {
        if (path.length === 0) {
            return this;
        }

        var _path5 = _toArray(path);

        const first = _path5[0],
              rest = _path5.slice(1);

        const nextNode = this.children[first];
        if (!nextNode) {
            return null;
        }
        return nextNode.subTrie(rest);
    }

    /**
     * Gets all elements that have a given prefix
     *
     * @param {Array<string>} path
     * @return {Array} All elements who match path as a prefix
     */
    search(path) {
        const sub = this.subTrie(path);
        return sub ? [...sub] : [];
    }

    /**
     * Determines if the trie contains a given path
     *
     * @param {Array<string>} path
     * @return {boolean} True if the path, `path`, exists in the trie
     */
    contains(path) {
        return this.search(path).length > 0;
    }

    *[Symbol.iterator]() {
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
            for (var _iterator = this.elements[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                let element = _step.value;

                yield element;
            }
        } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion && _iterator.return) {
                    _iterator.return();
                }
            } finally {
                if (_didIteratorError) {
                    throw _iteratorError;
                }
            }
        }

        for (let childName in this.children) {
            yield* this.children[childName][Symbol.iterator]();
        }
    }

    entriesIter(path = []) {
        const t = this;

        return {
            *[Symbol.iterator]() {
                var _iteratorNormalCompletion2 = true;
                var _didIteratorError2 = false;
                var _iteratorError2 = undefined;

                try {
                    for (var _iterator2 = t.elements[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                        let element = _step2.value;

                        yield [path, element];
                    }
                } catch (err) {
                    _didIteratorError2 = true;
                    _iteratorError2 = err;
                } finally {
                    try {
                        if (!_iteratorNormalCompletion2 && _iterator2.return) {
                            _iterator2.return();
                        }
                    } finally {
                        if (_didIteratorError2) {
                            throw _iteratorError2;
                        }
                    }
                }

                for (let childName in t.children) {
                    yield* t.children[childName].entriesIter(path.concat([childName]))[Symbol.iterator]();
                }
            }
        };
    }
}

module.exports = Node;